import logging

from fastapi.security.base import SecurityBase
from fastapi_azure_auth import SingleTenantAzureAuthorizationCodeBearer
from pydantic import BaseModel

from isar.config.settings import settings


class Token(BaseModel):
    access_token: str
    token_type: str


class NoSecurity(SecurityBase):
    def __init__(self) -> None:
        self.scheme_name = "No Security"


class Authenticator:
    def __init__(
        self,
        app_client_id: str = settings.APP_CLIENT_ID,
        tenant_id: str = settings.AZURE_TENANT_ID,
        openapi_client_id: str = settings.OPENAPI_CLIENT_ID,
        authentication_enabled: bool = settings.AUTHENTICATION_ENABLED,
    ) -> None:
        self.logger = logging.getLogger("api")
        self.app_client_id: str = app_client_id
        self.tenant_id: str = tenant_id
        self.openapi_client_id: str = openapi_client_id

        self.authentication_enabled: bool = authentication_enabled

        enabled_string = "enabled" if self.authentication_enabled else "disabled"
        self.logger.info(f"API authentication is {enabled_string}")

    def should_authenticate(self):
        return self.authentication_enabled

    def get_azure_scheme(self):
        return SingleTenantAzureAuthorizationCodeBearer(
            app_client_id=self.app_client_id,
            tenant_id=self.tenant_id,
            scopes={
                f"api://{self.app_client_id}/user_impersonation": "user_impersonation",
            },
        )

    def get_scheme(self):
        if self.should_authenticate():
            return self.get_azure_scheme()
        return NoSecurity

    async def load_config(self):
        """
        Load OpenID config on startup.
        """
        if self.should_authenticate():
            await self.get_azure_scheme().openid_config.load_config()
        else:
            pass



from typing import List

from isar.models.mission import Mission


def is_robot_capable_of_mission(
    mission: Mission, robot_capabilities: List[str]
) -> bool:
    return all(
        [
            step.type in robot_capabilities
            for task in mission.tasks
            for step in task.steps
        ]
    )



from abc import ABCMeta, abstractmethod
from typing import Sequence

from robot_interface.models.initialize import InitializeParams
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission import InspectionStep, Step, StepStatus


class RobotInterface(metaclass=ABCMeta):
    """Interface to communicate with robots."""

    @abstractmethod
    def initiate_step(self, step: Step) -> None:
        """Send a step to the robot and initiate the execution of the step

        Parameters
        ----------
        step : Step
            The step that should be initiated on the robot.

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the step is not initiated.

        """
        raise NotImplementedError

    @abstractmethod
    def step_status(self) -> StepStatus:
        """Gets the status of the currently active step on robot.

        Parameters
        ----------
        None

        Returns
        -------
        StepStatus
            Status of the execution of current step.

        Raises:
        ------
        RobotException
            If the step status can't be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        """Stops the execution of the current step and stops the movement of the robot.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the robot is not stopped.

        """
        raise NotImplementedError

    @abstractmethod
    def get_inspections(self, step: InspectionStep) -> Sequence[Inspection]:
        """Return the inspections connected to the given step.

        Parameters
        ----------
        step : Step

        Returns
        -------
        Sequence[InspectionResult]
            List containing all the inspection results connected to the given step.

        Raises
        ------
        RobotException
            If the inspection results can't be retrieved.

        """
        raise NotImplementedError

    @abstractmethod
    def initialize(self, params: InitializeParams) -> None:
        """Initializes the robot. The initialization needed is robot dependent and the function can
        be a simple return statement if no initialization is needed for the robot.

        Parameters
        ----------
        kwargs: Any

        Returns
        -------
        None

        Raises
        ------
        RobotException
            If the initialization failed

        """
        raise NotImplementedError



import logging
import os
from abc import ABCMeta, abstractmethod
from queue import Empty, Queue
from typing import Tuple

import backoff
from paho.mqtt import client as mqtt
from paho.mqtt.client import Client

from isar.config.settings import settings


class MqttClientInterface(metaclass=ABCMeta):
    @abstractmethod
    def publish(
        self, topic: str, payload: str, qos: int = 0, retain: bool = False
    ) -> None:
        """
        Parameters
        ----------
        topic : string
            MQTT topic to publish to
        payload : string
            Payload to send to publish on the topic
        qos : integer
            Quality of Service
        retain : boolean
            Retain on topic

        Returns
        -------

        """
        pass


class MqttPublisher(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.mqtt_queue: Queue = mqtt_queue

    def publish(
        self, topic: str, payload: str, qos: int = 0, retain: bool = False
    ) -> None:
        queue_message: Tuple[str, str, int, bool] = (topic, payload, qos, retain)
        self.mqtt_queue.put(queue_message)


class MqttClient(MqttClientInterface):
    def __init__(self, mqtt_queue: Queue) -> None:
        self.logger = logging.getLogger("mqtt_client")

        self.mqtt_queue: Queue = mqtt_queue

        username: str = settings.MQTT_USERNAME
        password: str = ""
        try:
            password = os.environ["ISAR_MQTT_PASSWORD"]
        except KeyError:
            self.logger.warning(
                "Failed to retrieve ISAR_MQTT_PASSWORD from environment. Attempting "
                "with empty string as password."
            )

        self.host: str = settings.MQTT_HOST
        self.port: int = settings.MQTT_PORT

        self.client: Client = Client()

        self.client.enable_logger(logger=self.logger)

        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect

        self.client.username_pw_set(username=username, password=password)

    def run(self) -> None:
        self.connect(host=self.host, port=self.port)
        self.client.loop_start()

        while True:
            if not self.client.is_connected():
                continue
            try:
                topic, payload, qos, retain = self.mqtt_queue.get(timeout=1)
            except Empty:
                continue

            self.publish(topic=topic, payload=payload, qos=qos, retain=retain)

    def on_connect(self, client, userdata, flags, rc):
        self.logger.info("Connection returned result: " + mqtt.connack_string(rc))

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.logger.warning("Unexpected disconnection from MQTT Broker")

    @staticmethod
    def _on_success(data: dict) -> None:
        logging.getLogger("mqtt_client").info("Connected to MQTT Broker")
        logging.getLogger("mqtt_client").debug(
            f"Elapsed time: {data['elapsed']}, Tries: {data['tries']}"
        )

    @staticmethod
    def _on_backoff(data: dict) -> None:
        logging.getLogger("mqtt_client").warning(
            f"Failed to connect, retrying in {data['wait']} seconds"
        )

    @staticmethod
    def _on_giveup(data: dict) -> None:
        logging.getLogger("mqtt_client").error(
            "Failed to connect to MQTT Broker within set backoff strategy."
        )

    @backoff.on_exception(
        backoff.expo,
        ConnectionRefusedError,
        max_time=300,
        on_success=_on_success,
        on_backoff=_on_backoff,
        on_giveup=_on_giveup,
    )
    def connect(self, host: str, port: int) -> None:
        self.logger.info("Attempting to connect to MQTT Broker")
        self.logger.debug(f"Host: {host}, Port: {port}")
        self.client.connect(host=host, port=port)

    def publish(self, topic: str, payload: str, qos: int = 0, retain: bool = False):
        self.logger.debug(f"Publishing message to topic: {topic}")
        self.client.publish(topic=topic, payload=payload, qos=qos, retain=retain)



import logging
from copy import deepcopy
from typing import Any, Optional

from alitra import Pose
from injector import inject

from isar.config.settings import settings
from isar.models.communication.message import StartMissionMessage
from isar.models.communication.queues import QueueIO, Queues, QueueTimeoutError
from isar.models.mission.mission import Mission
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.state_machine.states_enum import States


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles
    required thread communication through queues to the state machine.
    """

    @inject
    def __init__(self, queues: Queues, queue_timeout: int = settings.QUEUE_TIMEOUT):
        self.queues = queues
        self.queue_timeout: int = queue_timeout
        self.logger = logging.getLogger("api")

    def get_state(self) -> States:
        return self.queues.state.check()

    def start_mission(self, mission: Mission, initial_pose: Optional[Pose]) -> None:
        self._send_command(
            StartMissionMessage(mission=deepcopy(mission), initial_pose=initial_pose),
            self.queues.start_mission,
        )

    def pause_mission(self) -> None:
        self._send_command(True, self.queues.pause_mission)

    def resume_mission(self) -> None:
        self._send_command(True, self.queues.resume_mission)

    def stop_mission(self) -> None:
        self._send_command(True, self.queues.stop_mission)

    def _send_command(self, input: Any, queueio: QueueIO) -> Any:
        queueio.input.put(input)
        try:
            return QueueUtilities.check_queue(
                queueio.output,
                self.queue_timeout,
            )
        except QueueTimeoutError as e:
            QueueUtilities.clear_queue(queueio.input)
            self.logger.error("Timeout while communicating with state machine")
            raise e



import logging
from http import HTTPStatus
from queue import Empty
from typing import List, Optional

import numpy as np
from alitra import Frame, Orientation, Pose, Position
from fastapi import Body, Query, Response
from injector import inject
from requests import HTTPError

from isar.apis.models import ApiPose, StartMissionResponse
from isar.config.settings import robot_settings, settings
from isar.mission_planner.mission_planner_interface import (
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.mission_planner.mission_validator import is_robot_capable_of_mission
from isar.models.communication.queues import QueueTimeoutError
from isar.models.mission import Mission, Task
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from robot_interface.models.mission import DriveToPose


class SchedulingController:
    @inject
    def __init__(
        self,
        mission_planner: MissionPlannerInterface,
        scheduling_utilities: SchedulingUtilities,
        queue_timeout: int = settings.QUEUE_TIMEOUT,
    ):
        self.logger = logging.getLogger("api")
        self.mission_planner: MissionPlannerInterface = mission_planner
        self.scheduling_utilities: SchedulingUtilities = scheduling_utilities
        self.queue_timeout: int = queue_timeout

    def start_mission(
        self,
        response: Response,
        mission_id: int = Query(
            ...,
            alias="ID",
            title="Mission ID",
            description="ID-number for predefined mission",
        ),
        initial_pose: Optional[ApiPose] = Body(
            default=None,
            description="The starting point of the mission. Used for initial localization of robot",
            embed=True,
        ),
    ):
        self.logger.info("Received request to start new mission")
        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            self.logger.error(
                "Internal Server Error - Current state of state machine unknown"
            )
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        if state in [
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Monitor,
            States.Paused,
        ]:
            self.logger.warning("Conflict - Mission already in progress")
            response.status_code = HTTPStatus.CONFLICT.value
            return

        try:
            mission: Mission = self.mission_planner.get_mission(mission_id)
        except HTTPError as e:
            self.logger.error(e)
            response.status_code = e.response.status_code
            return
        except MissionPlannerError as e:
            self.logger.error(e)
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        robot_capable: bool = is_robot_capable_of_mission(
            mission=mission, robot_capabilities=robot_settings.CAPABILITIES
        )
        if not robot_capable:
            self.logger.error(
                "Bad Request - Robot is not capable of performing mission"
            )
            response.status_code = HTTPStatus.BAD_REQUEST.value
            return

        initial_pose_alitra: Optional[Pose]
        if initial_pose:
            initial_pose_alitra = Pose(
                position=Position(
                    x=initial_pose.x,
                    y=initial_pose.y,
                    z=initial_pose.z,
                    frame=Frame("asset"),
                ),
                orientation=Orientation.from_euler_array(
                    euler=np.array(
                        [initial_pose.roll, initial_pose.pitch, initial_pose.yaw]
                    ),
                    frame=Frame("asset"),
                ),
                frame=Frame("asset"),
            )
        else:
            initial_pose_alitra = None

        self.logger.info(f"Starting mission: {mission.id}")

        try:
            self.scheduling_utilities.start_mission(
                mission=mission, initial_pose=initial_pose_alitra
            )
            self.logger.info("OK - Mission successfully started")
        except QueueTimeoutError:
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value
            self.logger.error("Timeout - Failed to start mission")
            return
        return StartMissionResponse(**mission.api_response_dict())

    def pause_mission(self, response: Response):
        self.logger.info("Received request to pause current mission")

        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            self.logger.error(
                "Internal Server Error - Current state of state machine unknown"
            )
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        if state in [States.Idle, States.StopStep, States.Paused, States.Initialize]:
            self.logger.info("Conflict - Pause command received in invalid state")
            response.status_code = HTTPStatus.CONFLICT.value
            return

        try:
            self.scheduling_utilities.pause_mission()
            self.logger.info("OK - Mission successfully paused")
        except QueueTimeoutError:
            self.logger.error("Timeout - Failed to pause mission")
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value

    def resume_mission(self, response: Response):
        self.logger.info("Received request to resume current mission")
        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            self.logger.error(
                "Internal Server Error - Current state of state machine unknown"
            )
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        if state in [
            States.Idle,
            States.InitiateStep,
            States.Monitor,
            States.StopStep,
            States.Initialize,
        ]:
            self.logger.info("Conflict - Resume command received in invalid state")
            response.status_code = HTTPStatus.CONFLICT.value
            return

        try:
            self.scheduling_utilities.resume_mission()
            self.logger.info("OK - Mission successfully resumed")
        except QueueTimeoutError:
            self.logger.error("Timeout - Failed to resume mission")
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value

    def stop_mission(self, response: Response):
        self.logger.info("Received request to stop current mission")

        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            self.logger.error(
                "Internal Server Error - Current state of state machine unknown"
            )
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        if state in [States.Idle, States.Initialize]:
            self.logger.info("Conflict - Stop command received in invalid state")
            response.status_code = HTTPStatus.CONFLICT.value
            return

        try:
            self.scheduling_utilities.stop_mission()
            self.logger.info("OK - Mission successfully stopped")
        except QueueTimeoutError:
            self.logger.error("Timeout - Failed to stop mission")
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value

    def drive_to(
        self,
        response: Response,
        x: float = Query(
            ...,
            alias="x-value",
            description="The target x coordinate",
        ),
        y: float = Query(
            ...,
            alias="y-value",
            description="The target y coordinate",
        ),
        z: float = Query(
            ...,
            alias="z-value",
            description="The target z coordinate",
        ),
        q: List[float] = Query(
            [0, 0, 0, 1],
            alias="quaternion",
            description="The target orientation as a quaternion (x,y,z,w)",
        ),
    ):
        try:
            state: States = self.scheduling_utilities.get_state()
        except Empty:
            self.logger.error(
                "Internal Server Error - Current state of state machine unknown"
            )
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
            return

        if state in [
            States.Initialize,
            States.InitiateStep,
            States.StopStep,
            States.Monitor,
            States.Paused,
        ]:
            self.logger.info("Conflict - Drivt to command received in invalid state")
            response.status_code = HTTPStatus.CONFLICT.value
            return

        robot_frame: Frame = Frame("robot")
        position: Position = Position(x=x, y=y, z=z, frame=robot_frame)
        orientation: Orientation = Orientation(
            x=q[0], y=q[1], z=q[2], w=q[3], frame=robot_frame
        )
        pose: Pose = Pose(position=position, orientation=orientation, frame=robot_frame)

        step: DriveToPose = DriveToPose(pose=pose)
        mission: Mission = Mission(tasks=[Task(steps=[step])])

        try:
            self.scheduling_utilities.start_mission(mission=mission, initial_pose=None)
            self.logger.info("OK - Drive to successfully started")
        except QueueTimeoutError:
            self.logger.error("Timout - Failed to start drive to")
            response.status_code = HTTPStatus.REQUEST_TIMEOUT.value



from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel


class StepResponse(BaseModel):
    id: UUID
    type: str


class TaskResponse(BaseModel):
    id: UUID
    tag_id: Optional[str]
    steps: List[StepResponse]


class StartMissionResponse(BaseModel):
    id: Union[UUID, int, str, None]
    tasks: List[TaskResponse]


class ApiPose(BaseModel):
    x: float
    y: float
    z: float
    roll: float
    pitch: float
    yaw: float



from http import HTTPStatus
from typing import Optional, Tuple

import pytest

from isar.apis.security.authentication import Authenticator
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.mission_planner_interface import MissionPlannerError
from isar.models.communication.queues.queue_timeout_error import QueueTimeoutError
from isar.services.utilities.queue_utilities import QueueUtilities
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.states_enum import States
from tests.mocks.mission_definition import MockMissionDefinition


def mock_check_queue(was_mission_started, state_at_request):
    mock_return_1 = was_mission_started, state_at_request
    if was_mission_started or state_at_request != "idle":
        mock_return_2 = False
    else:
        mock_return_2 = True

    return [mock_return_1, mock_return_2]


class TestSchedulerRoutes:
    @pytest.mark.parametrize(
        "mission_id, mock_get_mission, expected_exception,"
        "mock_get_state, mock_start_mission_side_effect, expected_status_code",
        [
            (
                12345,
                MockMissionDefinition.default_mission,
                MissionPlannerError,
                States.Idle,
                None,
                HTTPStatus.INTERNAL_SERVER_ERROR,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Idle,
                QueueTimeoutError,
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Monitor,
                True,
                HTTPStatus.CONFLICT,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Idle,
                QueueTimeoutError,
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                1,
                MockMissionDefinition.default_mission,
                None,
                States.Idle,
                None,
                HTTPStatus.OK,
            ),
        ],
    )
    def test_start_mission(
        self,
        client,
        access_token,
        mocker,
        mission_id,
        mock_get_mission,
        expected_exception,
        mock_get_state,
        mock_start_mission_side_effect,
        expected_status_code,
    ):
        mocker.patch.object(
            LocalPlanner,
            "get_mission",
            return_value=mock_get_mission,
            side_effect=expected_exception,
        )

        mocker.patch.object(
            SchedulingUtilities,
            "get_state",
            return_value=mock_get_state,
        )

        mocker.patch.object(
            SchedulingUtilities,
            "start_mission",
            side_effect=mock_start_mission_side_effect,
        )

        response = client.post(
            f"schedule/start-mission?ID={mission_id}",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "state, mock_return, expected_status_code",
        [
            (
                States.Monitor,
                QueueTimeoutError(),
                HTTPStatus.REQUEST_TIMEOUT,
            ),
            (
                States.InitiateStep,
                None,
                HTTPStatus.OK,
            ),
            (
                States.Idle,
                None,
                HTTPStatus.CONFLICT,
            ),
        ],
    )
    def test_stop_mission(
        self,
        client,
        access_token,
        mocker,
        state,
        mock_return,
        expected_status_code,
    ):
        mocker.patch.object(SchedulingUtilities, "get_state", return_value=state)
        mocker.patch.object(
            SchedulingUtilities, "stop_mission", side_effect=mock_return
        )

        response = client.post(
            "/schedule/stop-mission",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "mock_start_mission, mock_get_state,expected_status_code,request_params",
        [
            (
                None,
                States.Idle,
                HTTPStatus.OK,
                {"x": 1, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
            (
                QueueTimeoutError,
                States.Idle,
                HTTPStatus.REQUEST_TIMEOUT,
                {"x": 1, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
            (
                None,
                States.Monitor,
                HTTPStatus.CONFLICT,
                {"x": 1, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
            (
                None,
                States.Idle,
                HTTPStatus.UNPROCESSABLE_ENTITY,
                {"x": None, "y": 1, "z": 1, "orientation": "0,0,0,1"},
            ),
        ],
    )
    def test_schedule_drive_to(
        self,
        client,
        access_token,
        mock_start_mission,
        mock_get_state,
        mocker,
        expected_status_code,
        request_params,
    ):
        mocker.patch.object(
            SchedulingUtilities, "get_state", return_value=mock_get_state
        )

        mocker.patch.object(
            SchedulingUtilities, "start_mission", side_effect=mock_start_mission
        )
        mocker.patch.object(Authenticator, "should_authenticate", return_value=False)

        query_string = (
            f"x-value={request_params['x']}&"
            f"y-value={request_params['y']}&"
            f"z-value={request_params['z']}&"
            f"quaternion={request_params['orientation'][0]}&"
            f"quaternion={request_params['orientation'][2]}&"
            f"quaternion={request_params['orientation'][4]}&"
            f"quaternion={request_params['orientation'][6]}"
        )

        response = client.post(
            f"/schedule/drive-to?{query_string}",
            headers={"Authorization": "Bearer {}".format(access_token)},
        )

        assert response.status_code == expected_status_code
